@@ -26,17 +26,11 @@
  *  this program is GPLed. Yes.
  */
 
-/*
- * TODO: - rewrite it as a PS parser instead of a ps parser
- *           DONE
- *
- *       - autodetect audio and video channels
- *           DONE
- */
-
 #include "vlms.h"
 #include "ps.h"
 
+static void CalculateCRC( unsigned char *p_begin, unsigned char *p_end );
+ 
 /******************************************************************************
  * ps_thread
  ******************************************************************************
@@ -579,35 +573,35 @@ void write_pat(ps_t *ps, unsigned char *ts)
 {
     int i;
 
-    ts[0] = 0x47; /* sync_byte */
-    ts[1] = 0x40;
-    ts[2] = 0x00; /* PID = 0x0000 */
-    ts[3] = 0x10 | (ps->pat_counter & 0x0f);
-    ts[4] = 0x00;
+    ts[0x00] = 0x47; /* sync_byte */
+    ts[0x01] = 0x40;
+    ts[0x02] = 0x00; /* PID = 0x0000 */
+    ts[0x03] = 0x10 | (ps->pat_counter & 0x0f);
+    ts[0x04] = 0x00; /* CRC calculation begins here */
     
-    ts[5] = 0x00; /* 0x00: Program association section */
+    ts[0x05] = 0x00; /* 0x00: Program association section */
 
-    ts[6] = 0xb0; /* */
-    ts[7] = 0x11; /* section_length = 0x011 */
+    ts[0x06] = 0xb0; /* */
+    ts[0x07] = 0x11; /* section_length = 0x011 */
 
-    ts[8] = 0x00;
-    ts[9] = 0xbb; /* TS id = 0x00b0 (what the vlc calls "Stream ID") */
+    ts[0x08] = 0x00;
+    ts[0x09] = 0xbb; /* TS id = 0x00b0 (what the vlc calls "Stream ID") */
 
-    ts[10] = 0xc1;
+    ts[0x0a] = 0xc1;
     /* section # and last section # */
-    ts[11] = ts[12] = 0x00;
+    ts[0x0b] = ts[0x0c] = 0x00;
 
     /* Network PID (useless) */
-    ts[13] = ts[14] = 0x00; ts[15] = 0xe0; ts[16] = 0x10;
+    ts[0x0d] = ts[0x0e] = 0x00; ts[0x0f] = 0xe0; ts[0x10] = 0x10;
 
     /* Program Map PID */
-    ts[17] = 0x03; ts[18] = 0xe8; ts[19] = 0xe0; ts[20] = 0x64;
+    ts[0x11] = 0x03; ts[0x12] = 0xe8; ts[0x13] = 0xe0; ts[0x14] = 0x64;
 
-    /* CRC */
-    ts[21] = 0x4d; ts[22] = 0x6a; ts[23] = 0x8b; ts[24] = 0x0f;
+    /* Put CRC in ts[0x15...0x18] */
+    CalculateCRC( ts + 0x05, ts + 0x15 );
 
     /* needed stuffing bytes */
-    for (i=25 ; i < 188 ; i++) ts[i]=0xff;
+    for (i=0x19 ; i < 188 ; i++) ts[i]=0xff;
 
     ps->sent_ts++;
 }
@@ -620,90 +614,176 @@ void write_pmt(ps_t *ps, unsigned char *ts)
 {
     int i;
 
-    ts[0] = 0x47; /* sync_byte */
-    ts[1] = 0x40;
-    ts[2] = 0x0064; /* PID = 0x0064 */
-    ts[3] = 0x10 | (ps->pmt_counter & 0x0f);
-    ts[4] = 0x00;
+    ts[0x00] = 0x47; /* sync_byte */
+    ts[0x01] = 0x40;
+    ts[0x02] = 0x0064; /* PID = 0x0064 */
+    ts[0x03] = 0x10 | (ps->pmt_counter & 0x0f);
+    ts[0x04] = 0x00; /* CRC calculation begins here */
 
-    ts[5] = 0x02; /* 0x02: Program map section */
+    ts[0x05] = 0x02; /* 0x02: Program map section */
 
-    ts[6] = 0xb0; /* */
-    ts[7] = 0x34; /* section_length = 0x034 */
+    ts[0x06] = 0xb0; /* */
+    ts[0x07] = 0x25; /* section_length = 0x025 */
 
-    ts[8] = 0x03;
-    ts[9] = 0xe8; /* prog number */
+    ts[0x08] = 0x03;
+    ts[0x09] = 0xe8; /* prog number */
 
-    ts[10] = 0xc1;
+    ts[0x0a] = 0xc1;
     /* section # and last section # */
-    ts[11] = ts[12] = 0x00;
+    ts[0x0b] = ts[0x0c] = 0x00;
 
     /* PCR PID */
-    ts[13] = 0xe0;
-    ts[14] = 0x20;
+    ts[0x0d] = 0xe0;
+    ts[0x0e] = 0x20;
 
     /* program_info_length == 0 */
-    ts[15] = 0xf0; ts[16] = 0x00;
+    ts[0x0f] = 0xf0; ts[0x10] = 0x00;
 
     /* Program Map / Video PID */
-    ts[17] = 0x02; /* stream type = video */
-    ts[18] = 0xe0; ts[19] = 0x20;
-    ts[20] = 0xf0; ts[21] = 0x09; /* es info length */
+    ts[0x11] = 0x02; /* stream type = video */
+    ts[0x12] = 0xe0; ts[0x13] = 0x20;
+    ts[0x14] = 0xf0; ts[0x15] = 0x09; /* es info length */
     /* useless info */
-    ts[22] = 0x07; ts[23] = 0x04; ts[24] = 0x08; ts[25] = 0x80; ts[26] = 0x24;
-    ts[27] = 0x02; ts[28] = 0x11; ts[29] = 0x01; ts[30] = 0xfe;
+    ts[0x16] = 0x07; ts[0x17] = 0x04; ts[0x18] = 0x08; ts[0x19] = 0x80;
+    ts[0x1a] = 0x24; ts[0x1b] = 0x02; ts[0x1c] = 0x11; ts[0x1d] = 0x01;
+    ts[0x1e] = 0xfe;
 
     switch ( ps->audio_type )
     {
     case REQUESTED_AC3 :
         /* ac3 */
-        ts[31] = 0x81; /* stream type = audio */
-        ts[32] = 0xe0; ts[33] = 0x80 + ps->audio_channel;
-        ts[34] = 0xf0; ts[35] = 0x00; /* es info length */
+        ts[0x1f] = 0x91; /* stream type = audio */
+        ts[0x20] = 0xe0; ts[0x21] = 0x80 + ps->audio_channel;
+        ts[0x22] = 0xf0; ts[0x23] = 0x00; /* es info length */
         break;
 
     case REQUESTED_MPEG :
         /* mpeg */
-        ts[31] = 0x04; /* stream type = audio */
-        ts[32] = 0xe0; ts[33] = 0x40 + ps->audio_channel;
-        ts[34] = 0xf0; ts[35] = 0x00; /* es info length */
+        ts[0x1f] = 0x04; /* stream type = audio */
+        ts[0x20] = 0xe0; ts[0x21] = 0x40 + ps->audio_channel;
+        ts[0x22] = 0xf0; ts[0x23] = 0x00; /* es info length */
         break;
 
     case REQUESTED_LPCM :
         /* LPCM audio */
-        ts[31] = 0x81;
-        ts[32] = 0xe0; ts[33] = 0xa0 + ps->audio_channel;
-        ts[34] = 0xf0; ts[35] = 0x00; /* es info  length */
+        ts[0x1f] = 0x93;
+        ts[0x20] = 0xe0; ts[0x21] = 0xa0 + ps->audio_channel;
+        ts[0x22] = 0xf0; ts[0x23] = 0x00; /* es info  length */
         break;
 
     default :
         /* No audio */
-        ts[31] = 0x00;
-        ts[35] = 0x00; /* es info  length */
+        ts[0x1f] = 0x00;
+        ts[0x23] = 0x00; /* es info  length */
     }
 
     /* Subtitles */
     if( ps->subtitles_channel == NO_SUBTITLES )
     {
-        ts[36] = 0x00;
-        ts[37] = 0x00; ts[38] = 0x00;
-        ts[39] = 0xf0; ts[40] = 0x00; /* es info length */
+        ts[0x24] = 0x00;
+        ts[0x25] = 0x00; ts[0x26] = 0x00;
+        ts[0x27] = 0xf0; ts[0x28] = 0x00; /* es info length */
     }
     else
     {
-        ts[36] = 0x82;
-        ts[37] = 0xe0; ts[38] = 0xa0 + ( ps->subtitles_channel );
-        ts[39] = 0xf0; ts[40] = 0x00; /* es info length */
+        ts[0x24] = 0x92;
+        ts[0x25] = 0xe0; ts[0x26] = 0xa0 + ( ps->subtitles_channel );
+        ts[0x27] = 0xf0; ts[0x28] = 0x00; /* es info length */
         //fprintf(stderr, "subtitle %.2x\n", 0xa0 + ( ps->subtitles_channel ) );
     }
 
-    /* CRC FIXME: not calculated yet*/
-    ts[41] = 0x96; ts[42] = 0x70; ts[43] = 0x0b; ts[44] = 0x7c;
+    /* Put CRC in ts[0x29...0x2c] */
+    CalculateCRC( ts + 0x05, ts + 0x29 );
 
     /* needed stuffing bytes */
-    for (i=45 ; i < 188 ; i++) ts[i]=0xff;
+    for (i=0x2d ; i < 188 ; i++) ts[i]=0xff;
 
     ps->sent_ts++;
 }
 
+static void CalculateCRC( unsigned char *p_begin, unsigned char *p_end )
+{
+    static unsigned int CRC32[256] =
+    {
+        0x00000000, 0x04c11db7, 0x09823b6e, 0x0d4326d9,
+        0x130476dc, 0x17c56b6b, 0x1a864db2, 0x1e475005,
+        0x2608edb8, 0x22c9f00f, 0x2f8ad6d6, 0x2b4bcb61,
+        0x350c9b64, 0x31cd86d3, 0x3c8ea00a, 0x384fbdbd,
+        0x4c11db70, 0x48d0c6c7, 0x4593e01e, 0x4152fda9,
+        0x5f15adac, 0x5bd4b01b, 0x569796c2, 0x52568b75,
+        0x6a1936c8, 0x6ed82b7f, 0x639b0da6, 0x675a1011,
+        0x791d4014, 0x7ddc5da3, 0x709f7b7a, 0x745e66cd,
+        0x9823b6e0, 0x9ce2ab57, 0x91a18d8e, 0x95609039,
+        0x8b27c03c, 0x8fe6dd8b, 0x82a5fb52, 0x8664e6e5,
+        0xbe2b5b58, 0xbaea46ef, 0xb7a96036, 0xb3687d81,
+        0xad2f2d84, 0xa9ee3033, 0xa4ad16ea, 0xa06c0b5d,
+        0xd4326d90, 0xd0f37027, 0xddb056fe, 0xd9714b49,
+        0xc7361b4c, 0xc3f706fb, 0xceb42022, 0xca753d95,
+        0xf23a8028, 0xf6fb9d9f, 0xfbb8bb46, 0xff79a6f1,
+        0xe13ef6f4, 0xe5ffeb43, 0xe8bccd9a, 0xec7dd02d,
+        0x34867077, 0x30476dc0, 0x3d044b19, 0x39c556ae,
+        0x278206ab, 0x23431b1c, 0x2e003dc5, 0x2ac12072,
+        0x128e9dcf, 0x164f8078, 0x1b0ca6a1, 0x1fcdbb16,
+        0x018aeb13, 0x054bf6a4, 0x0808d07d, 0x0cc9cdca,
+        0x7897ab07, 0x7c56b6b0, 0x71159069, 0x75d48dde,
+        0x6b93dddb, 0x6f52c06c, 0x6211e6b5, 0x66d0fb02,
+        0x5e9f46bf, 0x5a5e5b08, 0x571d7dd1, 0x53dc6066,
+        0x4d9b3063, 0x495a2dd4, 0x44190b0d, 0x40d816ba,
+        0xaca5c697, 0xa864db20, 0xa527fdf9, 0xa1e6e04e,
+        0xbfa1b04b, 0xbb60adfc, 0xb6238b25, 0xb2e29692,
+        0x8aad2b2f, 0x8e6c3698, 0x832f1041, 0x87ee0df6,
+        0x99a95df3, 0x9d684044, 0x902b669d, 0x94ea7b2a,
+        0xe0b41de7, 0xe4750050, 0xe9362689, 0xedf73b3e,
+        0xf3b06b3b, 0xf771768c, 0xfa325055, 0xfef34de2,
+        0xc6bcf05f, 0xc27dede8, 0xcf3ecb31, 0xcbffd686,
+        0xd5b88683, 0xd1799b34, 0xdc3abded, 0xd8fba05a,
+        0x690ce0ee, 0x6dcdfd59, 0x608edb80, 0x644fc637,
+        0x7a089632, 0x7ec98b85, 0x738aad5c, 0x774bb0eb,
+        0x4f040d56, 0x4bc510e1, 0x46863638, 0x42472b8f,
+        0x5c007b8a, 0x58c1663d, 0x558240e4, 0x51435d53,
+        0x251d3b9e, 0x21dc2629, 0x2c9f00f0, 0x285e1d47,
+        0x36194d42, 0x32d850f5, 0x3f9b762c, 0x3b5a6b9b,
+        0x0315d626, 0x07d4cb91, 0x0a97ed48, 0x0e56f0ff,
+        0x1011a0fa, 0x14d0bd4d, 0x19939b94, 0x1d528623,
+        0xf12f560e, 0xf5ee4bb9, 0xf8ad6d60, 0xfc6c70d7,
+        0xe22b20d2, 0xe6ea3d65, 0xeba91bbc, 0xef68060b,
+        0xd727bbb6, 0xd3e6a601, 0xdea580d8, 0xda649d6f,
+        0xc423cd6a, 0xc0e2d0dd, 0xcda1f604, 0xc960ebb3,
+        0xbd3e8d7e, 0xb9ff90c9, 0xb4bcb610, 0xb07daba7,
+        0xae3afba2, 0xaafbe615, 0xa7b8c0cc, 0xa379dd7b,
+        0x9b3660c6, 0x9ff77d71, 0x92b45ba8, 0x9675461f,
+        0x8832161a, 0x8cf30bad, 0x81b02d74, 0x857130c3,
+        0x5d8a9099, 0x594b8d2e, 0x5408abf7, 0x50c9b640,
+        0x4e8ee645, 0x4a4ffbf2, 0x470cdd2b, 0x43cdc09c,
+        0x7b827d21, 0x7f436096, 0x7200464f, 0x76c15bf8,
+        0x68860bfd, 0x6c47164a, 0x61043093, 0x65c52d24,
+        0x119b4be9, 0x155a565e, 0x18197087, 0x1cd86d30,
+        0x029f3d35, 0x065e2082, 0x0b1d065b, 0x0fdc1bec,
+        0x3793a651, 0x3352bbe6, 0x3e119d3f, 0x3ad08088,
+        0x2497d08d, 0x2056cd3a, 0x2d15ebe3, 0x29d4f654,
+        0xc5a92679, 0xc1683bce, 0xcc2b1d17, 0xc8ea00a0,
+        0xd6ad50a5, 0xd26c4d12, 0xdf2f6bcb, 0xdbee767c,
+        0xe3a1cbc1, 0xe760d676, 0xea23f0af, 0xeee2ed18,
+        0xf0a5bd1d, 0xf464a0aa, 0xf9278673, 0xfde69bc4,
+        0x89b8fd09, 0x8d79e0be, 0x803ac667, 0x84fbdbd0,
+        0x9abc8bd5, 0x9e7d9662, 0x933eb0bb, 0x97ffad0c,
+        0xafb010b1, 0xab710d06, 0xa6322bdf, 0xa2f33668,
+        0xbcb4666d, 0xb8757bda, 0xb5365d03, 0xb1f740b4
+    };
+
+    unsigned int i_crc = 0xffffffff;
+
+    /* Calculate the CRC */
+    while( p_begin < p_end )
+    {
+        i_crc = (i_crc<<8) ^ CRC32[ (i_crc>>24) ^ ((unsigned int)*p_begin) ];
+        p_begin++;
+    }
+
+    /* Store it after the data */
+    p_end[0] = (i_crc >> 24) & 0xff;
+    p_end[1] = (i_crc >> 16) & 0xff;
+    p_end[2] = (i_crc >>  8) & 0xff;
+    p_end[3] = (i_crc >>  0) & 0xff;
+}
 
